from .node import Node 
from .tree import BaseTree
import warnings
import time
import random
import numpy as np
import os 
import subprocess as sp
"""
author: Haotian
created_at: 12/4/2021
description: utils for trees
"""

class TraversalGenerator(object):
    """
        A generator class used for tree traversal
        Arguments:
            order: traversal order
        >>> generator = TraversalGenerator(order='post')
        >>> for node in generator(tree):
                # to do something
                pass
    """

    def __init__(self, order='post'):
        self.order = order
        self.iterator = None

    def __call__(self, tree, order='post'):
        # calling function
        valid_methods = {'pre': self._pre, 'in': self._in, 'post': self._post}
        assert order in valid_methods, "order should be in ['pre', 'mid', 'post']"
        method = valid_methods[order]
        iterator = iter(method(tree))
        return iterator

    def _pre(self, tree):
        # pre-order traverse
        node = tree.root
        traverse_nodes = []
        while len(traverse_nodes) != len(tree):
            while not node.is_leaf():
                if node.identifier not in traverse_nodes:
                    traverse_nodes.append(node.identifier)
                    yield node
                flag = False
                for child in node.get_children():
                    if child.identifier not in traverse_nodes:
                        flag = True
                        node = child
                        break
                if not flag: node = node.parent
            if node.identifier not in traverse_nodes:
                traverse_nodes.append(node.identifier)
                yield node
            node = node.parent
                

    def _in(self, tree):
        # mid-order traverse
        warnings.warn("no implementation currently")

    def _post(self, tree):
        node = tree.root
        traverse_nodes = []
        while len(traverse_nodes) != len(tree):
            while node.get_children():
                flag = True
                for child in node.get_children():
                    if child.identifier not in traverse_nodes:
                        node = child
                        flag = False
                        break
                if flag:
                    break
            traverse_nodes.append(node.identifier)
            yield node
            node = node.parent

    @property
    def order(self):
        return self._order

    @order.setter
    def order(self, method):
        valid_methods = {'pre': self._pre, 'in': self._in, 'post': self._post}
        if method in valid_methods:
            self._order = method
            self._method = valid_methods[method]
        else:
            raise Exception("order should be in ['pre', 'mid', 'post']")


def from_node(node: Node) -> BaseTree:
    """ Build a tree by directly setting a root """
    tree = BaseTree(root=node)
    return tree


def from_newick(newick: str) -> BaseTree:
    """ Build a tree according to a newick-format string """
    def _isvalid(s):
        checking_stack = []
        for ch in s:
            checking_stack.append(ch) if ch == '(' else None
            if ch == ')':
                if checking_stack:
                    checking_stack.pop()
                else:
                    return False
        return True if not checking_stack and ch == ';' else False

    def _next(i):
        stop_words = [',', ')', ';']
        if newick[i] in stop_words:
            return None, 0, i
        br = 0
        j = i + 1
        while newick[j] not in stop_words:
            j += 1
        if ':' in newick[i:j]:
            nid, br = newick[i:j].split(':')
        else:
            nid = newick[i:j]
        return nid.strip(), br, j

    newick = newick.strip()
    assert isinstance(newick, str), Exception('newick should be a string.')
    assert _isvalid(newick), Exception('invalid newick string.')
    nodes = []
    level, i, key = 0, 0, ''
    while not newick[i] == ';':
        if newick[i] in [',', ' ']:
            i += 1
            continue
        if newick[i] == '(':
            level += 1
            i += 1
            continue
        if newick[i] == ')':
            if not nodes:
                raise Exception('newick: bad parsing.')
            identifier, branch, end = _next(i + 1)
            node = Node(identifier=identifier, branch=branch)
            while nodes:
                if nodes[-1][1] == level:
                    child = nodes.pop()
                    node.add_child(child[0])
                    child[0].set_parent(node)
                else:
                    break
            level -= 1
            nodes.append((node, level))
            i = end
            continue
        identifier, branch, end = _next(i)
        node = Node(identifier=identifier, branch=branch)
        nodes.append((node, level))
        i = end

    root = nodes.pop()
    tree = from_node(root[0])
    return tree


# perturb the branch lengths for data without clock property
def perturb_tree_length(tree, min=.5, max=1.5, mode='mul'):
    tree = tree.copy()
    for node_str in tree.get_all_nodes():
        node = tree[node_str]
        if not node.is_root():
            try:
                if mode == 'mul':
                    rand = np.random.uniform(min, max)
                    node.branch = np.round(node.branch * rand, 4)
                if mode == 'add':
                    rand = np.random.rand()
                    node.branch = np.round(node.branch + rand, 4)
            except ValueError:
                raise Exception('branch is not a valid number')
    return tree


# relabel the leaves, this is usually used when there are different start indices
# only be using when labels are integers
# name_map should be a dictionary which is formatted as {'old_name': 'new_name'}
def relabel(tree, offset=0, name_map=None):
    tree = tree.copy()
    # try:
    if offset != 0 and name_map:
        raise Exception('error, both offset and name_map are specified')
    elif offset != 0:
        for leaf in tree.get_leaves():
            tree[leaf].name = str(int(tree[leaf].name)+offset)
    elif offset == 0 and name_map == None:
        pass
    else:
        for leaf in tree.get_leaves():
            name = tree[leaf].name
            if name in name_map:
                tree[leaf].name = name_map[name]
    # except Exception as e:
    #     print(e)
    #     exit()
    return tree


"""
    Construct a perfect phylogeny (no conflicts) (Dan Gusfield/ https://doi.org/10.1002/net.3230210104)
    Author: haotian Z
    Date: 12/27/22
"""
class BNode(Node):
    # extended Node class for recording mutations 
    def __init__(self, identifier=None, name=None, branch=0):
        super().__init__(identifier, name, branch)
        self.mutations = []

    def add_mutations(self, mutations):
        self.mutations.extend(mutations)


#  vanilla implementation takes O(nm^2) that simply check every pair of columns to see if they are disjoint or one includes the other one.
def check_no_conflict_vanilla(mat):
    nrow, ncol = mat.shape
    gametes = [[0,1], [1,0], [1,1]]
    for i in range(ncol):
        for j in range(ncol):
            flag = False
            for gamete in gametes:
                # print(gamete, mat[:, [i,j]])
                if gamete not in mat[:, [i,j]].tolist():
                    flag = True
                    break
            if not flag:
                return False
    return True


def remove_homozygous_columns(mat):
    mask = mat.sum(axis=0) > 0
    return mat[:, mask]


# useless one
def binary_number(arr):
    s = 0
    for i, ele in enumerate(arr[::-1]):
        if ele:
            s += 2**i
    return s 


# sort by binary numbers and remove duplicates
def rearrangement(mat):
    nrow, ncol = mat.shape
    binary_strs = []
    for i in range(ncol):
        binary_strs.append(''.join([str(val) for val in mat[:, i]]))
    sorted_index = sorted(range(ncol), key=lambda x:binary_strs[x], reverse=True)
    # build index and remove duplicates
    groups = {}
    group = []
    final_index = []
    for i in range(ncol):
        group.append(sorted_index[i])
        if i==ncol-1 or binary_strs[sorted_index[i]] != binary_strs[sorted_index[i+1]]:
            final_index.append(sorted_index[i])
            groups[sorted_index[i]] = group
            group = []
    return final_index, groups
    

# get a matrix for determining L(j)
def preprocess(mat):
    nrow, ncol = mat.shape
    indices, groups = rearrangement(mat)
    mat_ = mat[:, indices].copy()
    # for loop to check every entity with value 1
    pre_mat = np.zeros_like(mat_)
    for i in range(mat_.shape[0]):
        pre = 0
        for j in range(mat_.shape[1]):
            if mat_[i,j] == 1:
                pre_mat[i,j] = pre
                pre = j+1
    return mat_, pre_mat, indices, groups


#  Gusfield here did some sortings based on binary numbers then make an improvement to O(nm) [n: # of cells, m: # of columns]
def check_no_conflict(mat):
    nrow, ncol = mat.shape
    mat_, L, indices, groups = preprocess(mat)
    for j in range(L.shape[1]):
        maxx = L[:, j].max()
        for i in range(L.shape[0]):
            if mat_[i,j] == 1 and L[i,j] != maxx:
                return False
    return True
                    

# phylogeny construction from Gusfield's paper from a conflict-free matrix
def build_perfect_phylogeny(mat):
    mat = remove_homozygous_columns(mat)
    if not check_no_conflict(mat):
        raise Exception('This are some conflicts in the matrix provided.')
    mat_, L, indices, groups = preprocess(mat)
    L = np.max(L, axis=0)
    # step 1: build mutation tree
    root = BNode(identifier='root')
    mutation_nodes = {}
    for mutation in groups:
        node = BNode()
        # node = BNode(name=str(groups[mutation]))
        node.add_mutations(groups[mutation])
        mutation_nodes[mutation] = node
    for j in range(len(L)):
        if L[j] == 0:
            mutation = indices[j]
            root.add_child(mutation_nodes[mutation])
            mutation_nodes[mutation].set_parent(root)
        else:
            mutation_from = indices[L[j]-1]
            mutation_to = indices[j]
            node_from = mutation_nodes[mutation_from]
            node_to = mutation_nodes[mutation_to]
            node_from.add_child(node_to)
            node_to.set_parent(node_from)

    # step 2: add cells into the mutation tree
    nrow, ncol = mat_.shape
    for i in range(nrow):
        max_index = ncol-np.argmax(mat_[i][::-1])-1
        last_mutation_node = mutation_nodes[indices[max_index]]
        leaf_node = BNode(identifier=i)
        last_mutation_node.add_child(leaf_node)
        leaf_node.set_parent(last_mutation_node)
    
    # phylogeny = popgen.utils.from_node(root)
    # phylogeny.print()
    # step 3: tree prune
    for label in mutation_nodes:
        node = mutation_nodes[label]
        if len(node.get_children()) == 1:
            child = node.get_children()[0]
            node.parent.children.pop(node.identifier)
            node.parent._children.remove(node)
            node.parent.add_child(child)
            child.set_parent(node.parent)
            child.add_mutations(node.mutations)
    phylogeny = from_node(root)
    return phylogeny

def single_spr_move(tree, verbose=False):
    """
        Note: only for binary tree
        a spr move is to randomly select a node and swap its parent with another node
        the parent node should not be the root and the other node cannot be its descendant
        return a new tree

        1. allow the subtree directly under the root
    """
    
    while True:
        tree = tree.copy()
        nodes = list(tree.get_all_nodes()) # identifiers
        node = tree[random.choice(nodes)]
        parent = node.parent
        if node.is_root():
            continue
        sibling = node.get_siblings()[0]
        if not parent.is_root():
            grandparent = parent.parent
            grandparent.remove_child(parent)
            parent.set_parent(None)
            parent.remove_child(sibling)
            sibling.set_parent(None)
            grandparent.add_child(sibling)
            sibling.set_parent(grandparent)
        else:
            parent.remove_child(sibling)
            sibling.set_parent(None)
            tree = BaseTree(sibling)
        tree._update()
        candidates = list(tree.get_all_nodes().values())
        if len(candidates) == 1:
            continue
        break
    if verbose:
        print('spr remove: ', node.identifier)
    while True:
        new_sibling = random.choice(candidates)
        if new_sibling == sibling:
            continue
        if verbose:
            print('spr regraft: ', new_sibling.identifier)
        if new_sibling.is_root():
            parent.add_child(new_sibling)
            new_sibling.set_parent(parent)
            tree = BaseTree(parent)
        else:
            parent.add_child(new_sibling)
            new_sibling.parent.remove_child(new_sibling)
            new_sibling.parent.add_child(parent)
            parent.set_parent(new_sibling.parent)
            new_sibling.set_parent(parent)        
        break
    tree._update()
    return tree

# def single_spr_move(tree):
#     """
#         Note: only for binary tree
#         a spr move is to randomly select a node and swap its parent with another node
#         the parent node should not be the root and the other node cannot be its descendant
#         return a new tree
#     """
#     tree = tree.copy()
#     nodes = list(tree.get_all_nodes()) # identifiers
    
#     while True:
#         node = tree[random.choice(nodes)]
#         parent = node.parent
#         if node.is_root():
#             continue
#         if parent.is_root():
#             continue
#         # get all nodes that are not descendants of node
#         candidates = []
#         for idx in nodes:
#             n = tree[idx]
#             if n not in node.get_descendants() and n != node:
#                 candidates.append(n)
#         if not candidates:
#             continue
                
#         new_sibling = random.choice(candidates)
#         if new_sibling.is_root():
#             continue
#         if new_sibling == parent:
#             continue
#         # swap
#         # print('node', node.identifier)
#         # print('sib', new_sibling.identifier)  
#         grandparent = parent.parent
#         sibling = parent.get_children()[0]
#         grandparent.remove_child(parent)
#         parent.set_parent(None)
#         parent.remove_child(sibling)
#         grandparent.add_child(sibling)
#         sibling.set_parent(grandparent)
#         sibling_parent = new_sibling.parent
#         sibling_parent.remove_child(new_sibling)
#         new_sibling.set_parent(None)
#         parent.add_child(new_sibling)
#         new_sibling.set_parent(parent)
#         sibling_parent.add_child(parent)
#         parent.set_parent(sibling_parent)
#         break

#     tree._update()
#     return tree


def spr_move(tree, move):
    for _ in range(move):
        tree = single_spr_move(tree)
    return tree


def spr_distance_cpp(newick1, newick2, upper_bound=None, spr_exec_path='/home/haz19024/softwares/rspr/rspr'):
    """
    Computes the SPR distance between two vectors.
    """
    assert os.path.exists(spr_exec_path), "rSPR binary file not found."
    if not upper_bound:
        a = sp.run([spr_exec_path, "-bb", "-q"], input=f'{newick1}\n{newick2}'.encode(), stdout=sp.PIPE)
    else:
        a = sp.run([spr_exec_path, "-bb", "-q", "-split-approx", str(upper_bound)], input=f'{newick1}\n{newick2}'.encode(), stdout=sp.PIPE)
    result = a.stdout.decode().split('\n')[-2].split('=')[1]
    return float(result) 


def spr_distance(tree1, tree2):    
    newick1 = str(tree1)
    newick2 = str(tree2)
    return spr_distance_cpp(newick1, newick2)

def get_random_binary_tree(n_leave, start_index=0, random_branch=True):
    # create a random binary tree given the number of leaves
    nodes = [BNode(identifier=i) for i in range(start_index, n_leave+start_index)]
    while len(nodes) > 1:
        i = np.random.randint(0, len(nodes))
        j = np.random.randint(0, len(nodes))
        while j == i:
            j = np.random.randint(0, len(nodes))
        node = Node()
        node.add_child(nodes[i])
        node.add_child(nodes[j])
        if random_branch:
            nodes[i].branch = np.round(np.random.rand(), 4)
            nodes[j].branch = np.round(np.random.rand(), 4)
        nodes[i].set_parent(node)
        nodes[j].set_parent(node)
        nodes.pop(max(i,j))
        nodes.pop(min(i,j))
        nodes.append(node)
    root = nodes.pop()
    tree = from_node(root)
    return tree


def build_no_repeat_clade(g):
    clades = set()


def mutated_clade_dist(g1, g2):
    pass


# Here, we provide a Python interface for conveniently calling Scistree2.
import os
import sys
import multiprocessing as mp
import uuid
import numpy as np
import pandas as pd
import subprocess as sp
import importlib.resources
from .treeutils import *


def get_executable_path(provided_path=None):
    if provided_path:
        return provided_path
    binary_name = "scistree.exe" if sys.platform == "win32" else "scistree"
    return os.path.join(os.path.dirname(__file__), "bin", binary_name)


class ScisTree2:
    """
    Scistree2 Caller.

    Args:
        bin_path: Path to binary executable file.
        threads: Number of threads in use. Default uses all.
        nj: Call NJ only, serves when M is very big.
        spr: Enable SPR local search. Default is on.
        nni: Enable NNI local search. Default is off.
        iterative: Enable iterative optimization. Default is off.
        verbose: Show outputs.
    """

    def __init__(
        self, threads=-1, nj=False, spr=True, nni=False, max_iter=0, verbose=True
    ):
        self.bin_path = get_executable_path()
        self.nj = nj
        self.spr = spr
        self.nni = nni
        assert max_iter >= 0, "max_iter should be positive."
        self.max_iter = max_iter
        self.cmd = self.build_cmd(self.bin_path, threads, nj, nni, verbose)

    def build_cmd(self, bin_path, threads, nj, nni, verbose):
        if threads == -1:
            threads = mp.cpu_count()
        cmd = [bin_path, "-T", str(threads)]
        if verbose:
            cmd.append("-v")
        if nj:
            cmd.append("-n")
        if nni:
            cmd.append("-q")
        if self.max_iter:
            cmd.append("-s")
            cmd.append(str(self.max_iter))
        return cmd

    @staticmethod
    def write_to_scistree(genotype_matrix):
        """
        IO: Write the genotype matrix to a file.
        """
        nsite, ncell = genotype_matrix.shape
        prefix = uuid.uuid4()
        output = f"{prefix}.scistree.out"
        with open(output, "w") as out:
            out.write(f"HAPLOID\n")
            for i in range(nsite):
                for j in range(ncell):
                    prob = genotype_matrix[i, j]
                    out.write(f" {prob:.5f}")
                out.write("\n")
        return output

    @staticmethod
    def read_scistree_genotype(prefix):
        """
        Get the genotype matrix from the outputs of Scistree2.
        """
        geno_file = f"{prefix}.genos.imp"
        genotypes = []
        with open(geno_file, "r") as f:
            for line in f.readlines():
                if line.startswith("Site"):
                    line = line.strip()
                    genos = line.split("\t")[1].split()
                    genos = list(map(int, genos))
                    genotypes.append(genos)
        return np.array(genotypes)

    def bootstrap(self, tree, gp, num_bootstrap=100, num_site=-1):
        """
        Felsenstein’s tree bootstrp.
        """
        if num_site == -1:
            num_site = gp.nsite
        clades = {}
        for i in range(num_bootstrap):
            subgp = gp.subsample(n=num_site)
            t, geno, ml = self.infer(subgp)
            for split in t.get_splits():
                if split in clades:
                    clades[split] += 1
                else:
                    clades[split] = 1
        for node in tree.get_all_nodes():
            if tree[node].is_leaf() or tree[node].is_root():
                tree[node].branch_confidence = 1.0
            else:
                c = frozenset([leaf.name for leaf in tree[node].get_leaves()])
                if c in clades:
                    n = clades[c]
                else:
                    n = 0
                tree[node].branch_confidence = n / num_bootstrap
        return tree

    def infer(self, gp, verbose=False):
        """
        Run Scistree2 local search.

        Args:
            gp [scistree2.probability.GenotypeProbability]: Genotype probability matrix.
            verbose [bool]: Show logs.

        Returns:
            tree [scistree2.tree.BaseTree]: The optimal tree.
            imputed_genotype [pandas.DataFrame]: Imputed genotype. 0 as wild type, 1 as mutation.
            ml [float]: Log Likelihood of the optimal tree.
        """
        cell_names = gp.cell_names
        output = self.write_to_scistree(gp.probs)
        cmd = self.cmd + [f"{output}"]
        cmd = " ".join(cmd)
        try:
            res = (
                sp.run(cmd, shell=True, stdout=sp.PIPE, encoding="utf-8")
                .stdout.strip()
                .split("\n")
            )
            if verbose:
                print("\n".join(res))
            nwk = res[-2].split(":")[1].strip() + ";"
            if cell_names:
                t = relabel(
                    from_newick(nwk),
                    name_map={str(i + 1): name for i, name in enumerate(cell_names)},
                )
                nwk = t.output()
            imp_geno, ml, tree = evaluate(gp, nwk, return_tree=True)
            return tree, imp_geno, ml
        # clean output
        except Exception as e:
            print("scistree running failed.")
            if os.path.exists(output):
                os.remove(output)
            raise e
        finally:
            if os.path.exists(output):
                os.remove(output)
            if os.path.exists(f"{output}.genos.imp"):
                os.remove(f"{output}.genos.imp")


@staticmethod
def evaluate(gp, nwk, return_tree=False):
    """
    Evaluate a tree given genotype probabilities and return the imputated genotype.

    Args:
        gp [scistree2.probablity.GenotypeProbablity]: Genotype probability matrix.
        nwk [str]: Newick string of the optimal tree.
        return_tree: If return the optimal tree.

    Returns:
        imputed_genotype [pandas.DataFrame]: Imputed genotype. 0 as wild type, 1 as mutation.
        ml [float]: Log likelihood of the optimal tree.
        tree: Optinal
    """
    assert isinstance(nwk, str), "tree should be a newick string."
    cell_names = gp.cell_names
    site_names = gp.site_names
    tree = relabel(
        from_newick(nwk), name_map={name: str(i) for i, name in enumerate(cell_names)}
    )
    traveror = TraversalGenerator(order="post")
    max_mls = np.zeros(gp.probs.shape[0])  # - np.inf
    max_ml_nodes = [None] * gp.probs.shape[0]
    g = np.log(1 - gp.probs) - np.log(
        gp.probs
    )  # in log space to avoid numerical overflow
    for node in traveror(tree):
        if node.is_leaf():
            likelihood = g[:, int(node.name)]
        else:
            likelihood = (
                node.get_children()[0].likelihood + node.get_children()[1].likelihood
            )
        for i, l in enumerate(likelihood):
            if l > max_mls[i]:
                max_mls[i] = l
                max_ml_nodes[i] = node
        node.likelihood = likelihood
    max_mls += np.log(gp.probs).sum(axis=1)
    imputed_genotype = np.zeros_like(gp.probs, dtype=int)
    # print(imputed_genotype.sum(axis=-1))
    for i, ml_node in enumerate(max_ml_nodes):
        if ml_node:
            ml_node.add_mutation(site_names[i])
            inds = [int(leaf.name) for leaf in ml_node.get_leaves()]
            imputed_genotype[i, inds] = 1
    # rename leaves back to cell names
    tree = relabel(tree, name_map={str(i): name for i, name in enumerate(cell_names)})
    if return_tree:
        return (
            pd.DataFrame(imputed_genotype, index=gp.site_names, columns=gp.cell_names),
            sum(max_mls[max_mls != -np.inf]),
            tree,
        )
    return pd.DataFrame(
        imputed_genotype, index=gp.site_names, columns=gp.cell_names
    ), sum(max_mls[max_mls != -np.inf])



from collections import OrderedDict, defaultdict
from .node import Node
import pptree
import copy
import _pickle as cPickle


def deepcopy(obj):
    try:
        return cPickle.loads(cPickle.dumps(obj, -1))
    except Exception:
        return copy.deepcopy(obj)

class BaseTree(object):
    """
    A tree class, for extracting features from genealogical trees in the future.
    Arguments:
        args:
    # >>> tree = BaseTree();
    # >>> tree = popgen.utils.treeutils.from_newick('((1,2),(3,4));') # or
    """
    def __init__(self, root=None):
        self.root = root
        self._update()


    def __contains__(self, node):
        # if a node is in this tree
        if isinstance(node, Node):
            node = node.identifier
        if node in self._nodes:
            return True
        else:
            return False

    def __getitem__(self, identifier) -> Node:
        # get a node from node list
        return self._nodes[identifier]

    def __len__(self):
        # obtain number of nodes in a tree
        return len(self._nodes)
    
    def __eq__(self, tree):
        anchor = tree[tree.get_leaves()[0]].name
        return tree.get_splits(True, anchor) == self.get_splits(True, anchor)

    def __hash__(self):
        return hash(self.get_splits(True))

    def __str__(self):
        return self.output()

    def _update(self):
        if self.root is not None:
            self._nodes = self.root.get_descendants_dict()
            self._nodes[self.root.identifier] = self.root


    def create_node(self, identifier=None, name=None, parent=None) -> Node:
        node = Node(identifier=identifier, name=name)
        self.add_node(node, parent)
        return node

    def add_node(self, node, parent=None):
        if node.identifier in self._nodes:
            raise Exception('cannot add the node that has already been in the tree.')
        if parent is None:
            if self.root:
                raise Exception('root has already existed and parent cannot be none.')
            self.root = node
            self._nodes[node.identifier] = node
            # /* set root and make its level as 0 */
            node.set_parent(None)
            return
        pid = parent.identifier if isinstance(parent, Node) else parent
        if not pid in self._nodes:
            raise Exception('parent not found in this tree.')
        self._nodes[node.identifier] = node
        # /* link node with its parent */
        node.set_parent(self[pid])
        self[pid].add_child(node)
        return

    def get_all_nodes(self):
        return self._nodes

    def get_leaves(self, return_label=False):
        leaves = [node.name if return_label else node.identifier for node in self.root.get_leaves()]
        return leaves

    def get_splits(self, return_label=False, contains_leaf=None):
        def get_complement_split(split):
            return frozenset(filter(lambda x: x not in split, leaves))
        leaves = self.get_leaves(return_label=True)
        splits = set()
        for nid in self.get_all_nodes():
            # if it is root or a leaf which means trivial split, pass
            if not self._nodes[nid].is_leaf() and not self._nodes[nid].is_root():
                split = frozenset([node.name if return_label else node.identifier for node in self._nodes[nid].get_leaves()])
                if 1 < len(split) < len(leaves)-1 :
                    if contains_leaf and contains_leaf not in split:
                        splits.add(get_complement_split(split))
                    else:
                        splits.add(split)
        return frozenset(splits)

    def to_dict(self):
        # return a dict for the whole tree.
        pass

  
    def output(self, output_format='newick_sorted', branch_length_func=None, confidence_func=None):
        def _newick_unsorted(node, branch_length_func, confidence_func):
            if node.is_leaf():
                if branch_length_func and confidence_func:
                    b = branch_length_func(node)
                    c = confidence_func(node)
                    return f'({node.name}){c}:{b}'
                elif branch_length_func:
                    b = branch_length_func(node)
                    return f'{node.name}:{b}'
                elif confidence_func:
                    c = confidence_func(node)
                    return f'({node.name}{c})'
                return node.name
            fstr = '(' + ','.join([_newick_unsorted(child, branch_length_func, confidence_func)
                                   for child in node.get_children()]) + ')' +\
                                     (f'{confidence_func(node)}' if confidence_func else '') +\
                                    (f':{branch_length_func(node)}' if branch_length_func else '')
            return fstr

        def _newick_sorted(node, branch_length_func, confidence_func):
            if node.is_leaf():
                if branch_length_func and confidence_func:
                    b = branch_length_func(node)
                    c = confidence_func(node)
                    return f'({node.name}){c}:{b}'
                elif branch_length_func:
                    b = branch_length_func(node)
                    return f'{node.name}:{b}'
                elif confidence_func:
                    c = confidence_func(node)
                    return f'({node.name}{c})'
                return node.name
            newick_children = []
            for child in node.get_children():
                newick_children.append(_newick_sorted(child, branch_length_func, confidence_func))
            fstr = '(' + ','.join(sorted(newick_children)) + ')' + \
                    (f'{confidence_func(node)}' if confidence_func else '') +\
                    (f':{branch_length_func(node)}' if branch_length_func else '')
            return fstr
        
        def newick():
            return _newick_unsorted(self.root, branch_length_func, confidence_func) + ';'
        
        def newick_sorted():
            return _newick_sorted(self.root, branch_length_func, confidence_func) + ';'

        funcs = {'newick': newick, 'newick_sorted': newick_sorted}
        return funcs[output_format]()

    def draw(self, attr=None, **kwargs):
        if attr is not None:
            for node in self.get_all_nodes():
                self[node].__setattr__(attr, str(self[node].__getattribute__(attr)))
            pptree.print_tree(self.root, "_children", nameattr=attr, **kwargs)
        else:
            pptree.print_tree(self.root, "_children", **kwargs)

    def copy(self):
        return deepcopy(self)


import numpy as np

def tree_accuracy(tree1, tree2):
    count = 0
    splits1 = tree1.get_splits(return_label=True, contains_leaf='1')
    splits2 = tree2.get_splits(return_label=True, contains_leaf='1')
    for split in splits1:
        count += int(split in splits2)
    return count / len(splits1)

def split_accuracy(splits1, splits2):
    count = 0
    for split in splits1:
        count += int(split in splits2)
    return count / len(splits1)

def genotype_accuarcy(geno1, geno2):
    return np.mean(geno1 == geno2)


def is_covered(clade1, clade2):
    diff = clade1 - clade2
    if 0 in diff and 1 in diff and -1 in diff:
        return 0 # no relationship
    if 1 in diff:
        return 1 # clade1(snp1) happens before clade2(snp2)
    else:
        return -1 # clade2(snp2) happens before clade1(snp1)

def get_ancestor_descendant_pairs(geno):
    num_snp, _ = geno.shape
    mutations = {f's{_}': [] for _ in range(num_snp)}
    for i in range(num_snp):
        for j in range(i+1, num_snp):
            code = is_covered(geno[i], geno[j])
            if code == 1:
                mutations[f's{i}'].append(f's{j}')
            elif code == -1:
                mutations[f's{j}'].append(f's{i}')
    return mutations

def ancestor_descendant_error(mutation1, mutation2):
    count = 0
    total = 0
    for snp1 in mutation1:
        for snp2 in mutation1[snp1]:
            count += int(snp2 not in mutation2[snp1])
            total += 1
    # total = len(mutation1) * (len(mutation1) - 1) / 2
    return count / total

def different_lineage_error(mutation1, mutation2):
    count = 0
    total = 0
    muts = list(mutation1.keys())
    num_mut = len(muts)
    for i in range(num_mut):
        for j in range(i+1, num_mut):
            if muts[i] not in mutation1[muts[j]] and muts[j] not in mutation1[muts[i]]:
                count += int(muts[j] in mutation2[muts[i]]) + int(muts[i] in mutation2[muts[j]])
                total += 1
    return count / total


import numpy as np
import pandas as pd 
from .reader import read_vcf, read_csv

class GenotypeProbability():
    def __init__(self, probs, cell_names=None, site_names=None):
        self.probs = probs
        self.nsite, self.ncell = probs.shape
        self.shape = (self.nsite, self.ncell)
        if cell_names:
            self.cell_names = cell_names
        else:
            self.cell_names = [f'c{i}' for i in range(1, self.ncell+1)]
        if site_names:
            self.site_names = site_names
        else:
            self.site_names = [f's{i}' for i in range(1, self.nsite+1)]  

    def __str__(self):
        df = pd.DataFrame(data=np.round(self.probs, 2), columns=self.cell_names, index=self.site_names)
        # return f'{self.ncell} cells, {self.nsite} sites'
        return df.__str__()
    
    def save(self, out_name='probs.csv'):
        df = pd.DataFrame(data=np.round(self.probs, 4), columns=self.cell_names, index=self.site_names)
        df.to_csv(out_name)

    def copy(self):
        probs = self.probs.copy()
        cell_names = self.cell_names[:]
        site_names = self.site_names[:]
        return GenotypeProbability(probs, cell_names, site_names)

    def subsample(self, n, shuffle_taxa=True):
        assert n <= self.nsite, "sampling size is larger than self.nsite."
        gp = self.copy()
        inds_site= np.random.choice(np.arange(gp.nsite), n, replace=True)
        inds_cell = np.arange(gp.ncell)
        if shuffle_taxa:
            np.random.shuffle(inds_cell)
        probs = gp.probs[inds_site][:, inds_cell]
        cell_names = [gp.cell_names[i] for i in inds_cell]
        site_names = [gp.site_names[i] for i in inds_site]
        return GenotypeProbability(probs, cell_names, site_names) 



def from_probs(probs, cell_names=None, site_names=None, margin=1e-5):
    probs = np.clip(probs, a_min=margin, a_max=1-margin)
    return GenotypeProbability(probs, cell_names=cell_names, site_names=site_names)


def from_reads(reads, ado=0.2, seqerr=0.01, posterior=True, af=None, cell_names=None, site_names=None):
    probs = genotype_probability(reads, ado=ado, seqerr=seqerr, posterior=posterior, af=af)
    return from_probs(probs, cell_names=cell_names, site_names=site_names)


def from_vcf(vcf_path, ado=0.2, seqerr=0.01, posterior=True, af=None, key='AD'):
    # TODO: AF may exist!
    reads, cell_names, site_names = read_vcf(vcf_path, key=key)
    return from_reads(reads, ado=ado, seqerr=seqerr, posterior=posterior, af=af, cell_names=cell_names, site_names=site_names)


def from_csv(csv_path, source='probability', ado=0.2, seqerr=0.01, posterior=True, af=None):
    assert source in ['probability', 'read'], "source should be either 'probability' or 'read'."
    if source == 'probability':
        probs, cell_names, site_names = read_csv(csv_path, reads=False)
        return from_probs(probs, cell_names, site_names)
    else:
        reads, cell_names, site_names = read_csv(csv_path, reads=True)
        return from_reads(reads, ado=ado, seqerr=seqerr, posterior=posterior, af=af, cell_names=cell_names, site_names=site_names)


def concatenate_strings(val1, val2):
    return str(val1) + "|" + str(val2)


def get_mask(ref_counts, alt_counts):
    total_counts = ref_counts + alt_counts
    mask = total_counts == 0
    return mask


def genotype_probability(reads, ado=0.2, seqerr=0.01, posterior=True, af=None):
    ref_counts = reads[:, :, 0]
    alt_counts = reads[:, : ,1]
    mask = get_mask(ref_counts, alt_counts)    
    l00, l01, l11 = likelihood_GATK(ref_counts, alt_counts, ado, seqerr)
    if posterior:
        if af is None:
            af = allele_frequency(l00, l01, l11)
            prob = posterior_probability_GATK(l00, l01, l11, prior_ref=af)
        else:
            prob = posterior_probability_GATK(l00, l01, l11, prior_ref=af)
    else:
        prob = posterior_probability_GATK(l00, l01, l11)
    return np.ma.array(prob, mask=mask).filled(fill_value=0.5)


def allele_frequency(l00, l01, l11):
    """
    Calculate the allele frequency for each site.
    """
    ml_gt = np.argmax(np.concatenate([l00[:, :, np.newaxis], l01[:, :, np.newaxis], l11[:, :, np.newaxis]], axis=-1), axis=-1)
    af = np.mean(ml_gt, axis=-1) / 2
    af = af[:, np.newaxis]
    return 1 - af
    

def posterior_probability_GATK(l00, l01, l11, prior_ref=0.5, margin=1e-5):
    """
    Calculate the posterior probability of each SNV being a true positive.

    We use GATK likelihood with ADO included as described in CellCoal manual (https://dapogon.github.io/cellcoal/cellcoal.manual.v1.1.html#537_genotype_likelihoods)
        P(D|G={g_1, g_2}) = (1-ado)\prod_{i=1}^{r}P(b_i|G={g_1, g_2}) + 0.5*ado[\prod_{i=1}^{r}P(b_i|G={g_1}) + \prod_{i=1}^{r}P(b_i|G={g_2})]
    """
    g00 = prior_ref**2 * l00
    g01 = 2*(1 - prior_ref) * prior_ref * l01
    g11 = (1 - prior_ref)**2 * l11
    g = g00 / (g00 + g01 + g11)
    g = np.clip(np.round(g, 5), a_min=margin, a_max=1-margin)
    return g


def likelihood_GATK(ref_counts, alt_counts, ado=0.2, seqerr=0.01):
    # Q-phred score = -10 * log10(p)
    ref_counts = ref_counts.astype(float)
    alt_counts = alt_counts.astype(float)
    p00, p01, p10, p11 = np.log(1-seqerr), np.log(seqerr), np.log(seqerr), np.log(1-seqerr)
    l00 = np.exp(ref_counts * p00 + alt_counts * p01)
    l01 = (1 - ado) * np.exp(ref_counts * np.log(0.5 * np.exp(p00) + 0.5 * np.exp(p10)) + alt_counts * np.log(0.5 * np.exp(p01) + 0.5 * np.exp(p11))) \
            + (0.5 * ado) * (np.exp(ref_counts * p00 + alt_counts * p10) + np.exp(ref_counts * p10 + alt_counts * p11))
    l11 = np.exp(ref_counts*p10 + alt_counts*p11)
    return l00, l01, l11




from collections import OrderedDict
import os
import binascii
import time

class Node(object):
    """
    A node class, was implemented for constructing genealogical trees.
    Arguments:
        args:
    >>> node = Node()
    >>> node = Node(1) # or
    >>> node = Node(identifier=1, name='ken', branch=10) # or
    """
    def __init__(self, identifier=None, name=None, branch=None):
        self.identifier = identifier
        if name is None:
            self.name = str(self.identifier)
        else:
            self.name = name
        self.parent = {}
        self.branch = branch
        self.mutations = []
        self.children = OrderedDict()
        self._children = []  # using for printing tree

    def __eq__(self, node):
        if node.identifier == self.identifier:
            return True
        return False
        
    def __hash__(self):
        return hash(self.identifier)

    @property
    def identifier(self):
        return self._identifier

    @identifier.setter
    def identifier(self, nid):
        if nid is None or nid == '':
            nid = str(binascii.hexlify(os.urandom(5)).decode('utf-8'))
        self._identifier = nid

    @property
    def branch(self):
        return self._branch

    @branch.setter
    def branch(self, value):
        if self.parent is None:
            raise Exception('node has no parent, branch cannot be assigned.')
        try:
            self._branch = float(value)
        except (ValueError, TypeError):
            # branch may not be a number, now we can assign any object attached to the branch
            # to be efficiently used on phylogeny and mutation tree
            self._branch = value 

    def is_root(self):
        return self.get_level() == 0

    def is_leaf(self):
        return len(self.children) == 0

    def get_ancestors(self):
        parents = []
        node = self.parent
        while node:
            parents.append(node)
            node = node.parent
        return parents

    def get_descendants_dict(self):
        # simple bfs for searching descendants
        descendants = {}
        queue = [self]
        while queue:
            node = queue.pop(0)
            for child in node.get_children():
                descendants.update({child.identifier: child})
                queue.append(child)
        return descendants

    def get_siblings(self):
        if not self.parent:
            return None
        siblings = []
        for sib in self.parent.get_children():
            if sib.identifier != self.identifier:
                siblings.append(sib)
        return siblings

    def get_descendants(self):
        # simple bfs for searching descendants
        descendants = []
        queue = [self]
        while queue:
            node = queue.pop(0)
            for child in node.get_children():
                descendants.append(child)
                queue.append(child)
        return descendants

    def add_child(self, node):
        """
        Adding child to a specific node. Potentially running slow due to inner traversal for ensuring no conflicts
        happened. (should be optimized in the future)
        """
        if node.identifier in [d.identifier for d in self.get_descendants()]:
            raise Exception('node %s has already been added.' % node.identifier)
        if node.identifier in [a.identifier for a in self.get_ancestors()]:
            raise Exception('parent %s cannot be added as a child.' % node.identifier)
        self.children[node.identifier] = node
        self._children.append(node)

    def add_mutation(self, site):
        if not site in self.mutations:
            self.mutations.append(site)

    def remove_child(self, node):
        if node.identifier not in self.children:
            raise Exception('node %s is not a child of %s.' % (node.identifier, self.identifier))
        del self.children[node.identifier]
        self._children.remove(node) 

    def get_children(self):
        return self._children

    def get_leaves(self):
        if self.is_leaf():
            return [self]
        leaves = []
        for node in self.get_descendants():
            if node.is_leaf():
                leaves.append(node)
        return leaves

    def get_level(self):
        level = 0
        node = self.parent
        while node:
            node = node.parent
            level += 1
        return level

    def set_parent(self, parent):
        if parent is None:
            self.parent = None
        else:
            self.parent = parent

    def set_branch(self, value):
        self.branch = value


if __name__ == '__main__':
    st = time.time()
    pre = Node()
    for i in range(10000):
        node = Node()
        pre.add_child(node)
        node.set_parent(pre)
        pre = node

    print(time.time()- st)



from setuptools import setup, find_packages, Distribution
from setuptools.command.build_py import build_py
import subprocess
import shutil
import os
import sys
from pathlib import Path


PACKAGE_NAME = 'scistree2'
# PROJECT_ROOT = str(Path(__file__).parent.resolve())
PACKAGE_DIR = 'scistree2'
SRC_DIR = "src"
CPP_DIR = "src"
# CPP_DIR = f'{PROJECT_ROOT}/src'
# PACKAGE_DIR = f'{PROJECT_ROOT}/scistree2'


class BinaryDistribution(Distribution):
    def has_ext_modules(self):
        return True  # <--- The magic line. Forces "platlib" (platform-specific) wheel.
    

# class BuildWithMake(build_py):
#     def run(self):
#         self.build_cpp_with_make()
#         super().run()
    
#     def build_cpp_with_make(self):
#         print("Building C++ project with make...")
#         try:
#             subprocess.run(['make'], cwd=CPP_DIR, check=True)
#         except subprocess.CalledProcessError as e:
#             print("Make failed.")
#             raise e
        
#         bin_dir = os.path.join(PACKAGE_DIR, 'bin')
#         os.makedirs(bin_dir, exist_ok=True)
#         shutil.copy(
#             os.path.join(CPP_DIR, 'scistree'),
#             os.path.join(bin_dir, 'scistree')
#         )
class CustomBuild(build_py):
    def run(self):
        if os.environ.get("CIBUILDWHEEL") == "1":
            print("--- Detected CIBUILDWHEEL: Skipping setup.py compilation (relying on pre-built binary) ---")
        else:
            # Local install (pip install .): We must compile.
            self.compile_cpp()
        super().run()

    def compile_cpp(self):
        print(f"--- Compiling C++ extensions for {sys.platform} ---")
        
        bin_dir = os.path.join(PACKAGE_DIR, "bin")
        os.makedirs(bin_dir, exist_ok=True)

        if sys.platform == "win32":
            # Windows: Use nmake
            build_cmd = "nmake /f Makefile.win"
            clean_cmd = "nmake /f Makefile.win clean"
            binary_name_src = "scistree.exe"
            binary_name_dest = "scistree.exe"
        else:
            # Linux/Mac: Use standard make
            build_cmd = "make"
            clean_cmd = "make clean"
            binary_name_src = "scistree"
            binary_name_dest = "scistree"

        try:
            # Run clean then build inside the 'src' directory
            subprocess.check_call(clean_cmd, cwd=SRC_DIR, shell=True)
            subprocess.check_call(build_cmd, cwd=SRC_DIR, shell=True)
        except subprocess.CalledProcessError as e:
            print("Error during compilation:", e)
            sys.exit(1)
        except FileNotFoundError:
            print(f"Error: Build tool not found. Ensure '{build_cmd[0]}' is in PATH.")
            sys.exit(1)

        src_path = os.path.join(SRC_DIR, binary_name_src)
        dest_path = os.path.join(bin_dir, binary_name_dest)
        
        if os.path.exists(src_path):
            print(f"Moving {src_path} -> {dest_path}")
            shutil.copy(src_path, dest_path) 
        else:
            print(f"Error: Compiled binary {binary_name_src} not found in {SRC_DIR}")
            sys.exit(1)

setup(
    name=PACKAGE_NAME,
    packages=[PACKAGE_DIR],
    cmdclass={'build_py': CustomBuild},
    install_requires=[
        'numpy',
        'pptree',
        'phytreeviz'
    ],
    package_data={
        PACKAGE_NAME: ['bin/*'],  # Corrected path
    },
    include_package_data=False,
    distclass=BinaryDistribution,
)



import os
# from ... import popgen
import numpy as np 
from . import metric
from . import probability
from . import treeutils as util
from .scistree import ScisTree2, evaluate
from .reader import read_vcf



### version updates ###
# __version__ = "0.1.0" # original submission
# __version__ = "0.2.0" # add ete3 tree, add mutation profile at branches.
# __version__ = "0.3.0" # add bootstrapping.
# __version__ = "0.4.0" # to pypi.
__version__ = "0.5.0" # add win support.


import os 
import numpy as np
import pandas as pd 

def read_vcf(vcf_filepath, key='AD'):
    """
    Reads a VCF file and extracts the Allelic Depth (AD) for each sample at each site.

    Args:
        vcf_filepath (str): The path to the VCF file.

    Returns:
        tuple[list[str], list[list[tuple[int | None, int | None]]]]:
            A tuple containing:
            - A matrix (list of lists) where each inner list represents a variant site,
              and each element in the inner list is a tuple (ref_reads, alt_reads)
              for a given sample. If AD is not available or malformed for a sample
              at a site, (None, None) will be used.
            Lines where the 'AD' field is not present in the FORMAT column will be skipped.
            - A list of sample names.
            - A list of site names.
    """
    sample_names = []
    ad_matrix = []

    assert os.path.exists(vcf_filepath), (f"Error: VCF file not found at '{vcf_filepath}'")

    with open(vcf_filepath, 'r') as f:
        site_names = []
        for line in f:
            line = line.strip()
            if not line:
                continue
            if line.startswith('#CHROM'):
                parts = line.split('\t')
                sample_names = parts[9:]
                continue
            if line.startswith('#'):
                continue
            parts = line.split('\t')
            if len(parts) < 9:
                print(f"Warning: Skipping malformed line (too few columns): {line}")
                continue
            format_str = parts[8]
            format_fields = format_str.split(':')
            try:
                ad_index = format_fields.index(key)
            except ValueError:
                print(f"Info: Skipping line because {key} not found in FORMAT: CHR[{parts[0]}], POS[{parts[1]}]")
                continue
            chrom = parts[0]
            pos = parts[1]
            current_site_name = f"{chrom}:{pos}"
            site_names.append(current_site_name)
            site_ad_data = []
            for sample_gt_str in parts[9:]:
                ref_reads, alt_reads = 0, 0
                gt_fields = sample_gt_str.split(':')
                if ad_index < len(gt_fields):
                    ad_value_str = gt_fields[ad_index]
                    if ad_value_str != '.' and ',' in ad_value_str:
                        try:
                            ad_parts = ad_value_str.split(',')
                            ref_reads = int(ad_parts[0])
                            alt_reads = int(ad_parts[1])
                        except (ValueError, IndexError):
                            pass
                site_ad_data.append((ref_reads, alt_reads))
            ad_matrix.append(site_ad_data)
    return np.array(ad_matrix), sample_names, site_names


def convert_2d_string_array_to_3d(input_2d_array):
    """
    Converts a 2D array (list of lists) of strings formatted as 'x|y'
    back into a 3D NumPy array of shape (rows, cols, 2).

    Args:
        input_2d_array (list): A 2D list of strings, where each string is 'x|y'.

    Returns:
        np.ndarray: A 3D NumPy array with the last dimension being of size 2.
                    Returns an empty array if the input list is empty.
    """
    num_rows = len(input_2d_array)
    num_cols = len(input_2d_array[0]) if num_rows > 0 else 0
    result_3d_array = np.empty((num_rows, num_cols, 2), dtype=object)

    for i in range(num_rows):
        for j in range(num_cols):
            parts = input_2d_array[i][j].split('|')
            if len(parts) == 2:
                try:
                    x_val = int(parts[0])
                    y_val = int(parts[1])
                    result_3d_array[i, j] = [x_val, y_val]
                except ValueError:
                    print(f"Warning: Could not convert '{input_2d_array[i][j]}' to integers. Storing as original string parts.")
                    result_3d_array[i, j] = parts
            else:
                print(f"Warning: Unexpected format for string '{input_2d_array[i][j]}'. Expected 'x|y'.")
                result_3d_array[i, j] = [None, None]
    return result_3d_array


def read_csv(csv_filepath, reads=False):
    """
    Reads a CSV file.
    Args:
        csv_filepath (str): The path to the CSV file.

    Returns:
        tuple[list[str], list[list[tuple[int | None, int | None]]]]:
            A tuple containing:
            - A matrix (list of lists) where each inner list represents a variant site,
              and each element in the inner list is a tuple (ref_reads, alt_reads).
            - A list of sample names.
            - A list of site names.
    """
    assert os.path.exists(csv_filepath), (f"Error: CSV file not found at '{csv_filepath}'")
    df = pd.read_csv(csv_filepath, index_col=0)
    sample_names = df.columns.to_list()
    site_names = df.index.to_list()
    return convert_2d_string_array_to_3d(df.values) if reads else df.values, sample_names, site_names



if __name__ == "__main__":
    a = read_csv('../tutorials/data/toy_probs.csv', reads=False)
    print(a)